-- =============================================================================
-- Copyright © 2016-2017 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- Author: Farnoud Farahmand
-- =============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use work.ACORN_pkg.all;
use work.AEAD_pkg.all;
use IEEE.math_real."ceil";
use IEEE.math_real."log2";
use work.CAESAR_LWAPI_pkg.all;
use work.design_pkg.all;

entity ACORN_Control is
    port(
        --! Global
        clk             : in  std_logic;
        rst             : in  std_logic;
        --! PreProcessor (controls)
        key_ready       : out std_logic;
        key_valid       : in  std_logic;
        key_update      : in  std_logic;
        decrypt         : in  std_logic;
        bdi_ready       : out std_logic;
        bdi_valid       : in  std_logic;
        bdi_type        : in  std_logic_vector(4                -1 downto 0);
        bdi_eot         : in  std_logic;
        bdi_eoi         : in  std_logic;
        --! PostProcessor
        bdo             : out std_logic_vector(8      -1 downto 0);
        bdo_valid       : out std_logic;
        bdo_ready       : in  std_logic;
        msg_auth        : out std_logic;
        msg_auth_valid  : out std_logic;
        msg_auth_ready  : in  std_logic;
		bdo_valid_bytes : out  STD_LOGIC_VECTOR (PWdiv8   -1 downto 0);
        bdo_type        : out  STD_LOGIC_VECTOR (4       -1 downto 0);

        --! Control
        en_KeyReg       : out  std_logic;
        L_KeyReg        : out  std_logic;
        rst_KeyReg      : out  std_logic;
        en_StateReg     : out  std_logic;
        rst_StateReg    : out  std_logic;
        ca_set          : out  std_logic;
        cb_set          : out  std_logic;
        sel_LoadKey     : out  std_logic;
        is_final        : out  std_logic;
        sel_decrypt     : out  std_logic;
        tag_match       : in   std_logic;
        sel_M           : out  std_logic_vector(2 downto 0);
        end_of_block    : out std_logic
    );

end entity ACORN_Control;

architecture behavioral of ACORN_Control is

    type t_state is(
                    S_RESET, S_WAIT_START, S_WAIT_KEY, S_INIT_KEY, S_INIT_NPUB,
                    S_INIT_KEY2, S_PROC_AD, S_PROC_ADPAD1, S_PROC_ADPAD0,
                    S_PROC_PT, S_PROC_DATPAD1, S_PROC_DATPAD0, S_FINAL,
                    S_WAIT_MSG_AUTH
                );
    signal state       : t_state;
    signal state_next  : t_state;
    signal count_r     : std_logic_vector(7 downto 0);
    signal count_next  : std_logic_vector(7 downto 0);
    signal eoi_r       : std_logic;
    signal eoi_next    : std_logic;
    signal auth_fail_next  : std_logic;
    signal auth_fail_r     : std_logic;
    signal decrypt_r       : std_logic;
    signal en_decrypt      : std_logic;

begin

    p_fsm: process(clk)
    begin
        if rising_edge(clk) then
            if (rst = '1') then
                state <= S_RESET;
            else
                state       <= state_next;
            end if;
            count_r         <= count_next;
            eoi_r           <= eoi_next;
            auth_fail_r     <= auth_fail_next;
            if en_decrypt='1' then
                decrypt_r   <= decrypt;
            end if;
        end if;
    end process;

    p_comb: process(state, count_r, bdi_valid, key_valid, bdi_eoi, eoi_r, bdi_type,
                    bdi_eot, bdo_ready, decrypt_r, tag_match, auth_fail_r, msg_auth_ready)
    begin
        --! Default values
        state_next  <= state;
        count_next  <= count_r;
        eoi_next    <= eoi_r;
        auth_fail_next    <= auth_fail_r;

        en_KeyReg       <= '0';
        L_KeyReg        <= '0';
        rst_KeyReg      <= '0';
        en_StateReg     <= '0';
        rst_StateReg    <= '0';
        ca_set          <= '0';
        cb_set          <= '0';
        sel_LoadKey     <= '0';
        is_final        <= '0';
        sel_decrypt     <= '0';
        sel_M           <= "000";

        --! External
        key_ready      <= '0';
        bdi_ready      <= '0';
        bdo_valid      <= '0';
        msg_auth       <= '0';
        msg_auth_valid <= '0';
        en_decrypt     <= '0';
        end_of_block   <= '0';

		bdo_valid_bytes<=(others=>'1');

        case state is

            when S_RESET =>
                state_next		<= S_WAIT_START;
                count_next	 	<= (others => '0');
				rst_StateReg	<= '1';
                auth_fail_next  <= '0';

            when S_WAIT_START =>
                -- if (bdi_valid = '1') then
                    -- if (key_update = '1') then
                        -- state_next <= S_WAIT_KEY;
                    -- else
                        -- count_next  <= (others => '0');
                        -- ca_set      <= '1';
                        -- cb_set      <= '1';
						-- rst_StateReg	<= '1';
                        -- state_next      <= S_INIT_KEY;
                    -- end if;
                -- end if;
                --state_next <= S_WAIT_KEY;
                if key_valid='1' then
                    state_next <= S_WAIT_KEY;
                elsif bdi_valid='1' then
                    count_next  <= (others => '0');
                    ca_set      <= '1';
                    cb_set      <= '1';
                    rst_StateReg	<= '1';
                    state_next <= S_INIT_KEY;
                else
                    state_next      <= S_WAIT_START;
                end if;

            when S_WAIT_KEY =>
                key_ready   <= '1';
                sel_LoadKey <= '1';
                if (key_valid = '1') then
                    en_KeyReg   <= '1';
                    if (count_r = PdivM-1) then
                        count_next  <= (others => '0');
                        state_next  <= S_INIT_KEY;
                    else
                        count_next  <= count_r + 1;
                    end if;
                end if;

            when S_INIT_KEY =>
                en_KeyReg   <= '1';
                en_StateReg <= '1';
				ca_set      <= '1';
				cb_set      <= '1';
                if (count_r = PdivM-1) then
					count_next  <= (others => '0');
					state_next  <= S_INIT_NPUB;
				else
					count_next  <= count_r + 1;
				end if;

			when S_INIT_NPUB =>
				sel_M		<= "010";
				ca_set      <= '1';
				cb_set      <= '1';
				bdi_ready   <= '1';
				if (bdi_valid = '1') then
					en_StateReg		<= '1';
					if (count_r = PdivM-1) then
                        eoi_next <= bdi_eoi;
                        en_decrypt  <= '1';
						count_next  <= (others => '0');
						state_next  <= S_INIT_KEY2;
					else
						count_next  <= count_r + 1;
					end if;
                end if;

			when S_INIT_KEY2 =>
				ca_set      <= '1';
				cb_set      <= '1';
				if (count_r = 0) then
					sel_M		<= "001";
					en_StateReg <= '1';
					en_KeyReg	<= '1';
					count_next  <= count_r + 1;
				elsif (count_r <= 191) then
					sel_M		<= "000";
					en_StateReg <= '1';
					en_KeyReg	<= '1';
					count_next  <= count_r + 1;
				else
					count_next  <= (others => '0');
                    if (eoi_r = '1') then
                        state_next  <= S_PROC_ADPAD1;
                    else
                        state_next  <= S_PROC_AD;
                    end if;
				end if;

			when S_PROC_AD =>
                if (bdi_valid = '1') then
                    if (bdi_type = HDR_AD) then
                        bdi_ready   <= '1';
                        ca_set      <= '1';
                        cb_set      <= '1';
                        sel_M		<= "010";
                        en_StateReg <= '1';
                        if (bdi_eot = '1') then
                            eoi_next <= bdi_eoi;
                            state_next <= S_PROC_ADPAD1;
                        end if;
                    else
                        ca_set      <= '1';
                        cb_set      <= '1';
                        sel_M		<= "011";
                        en_StateReg <= '1';
                        count_next  <= count_r + 1;
                        state_next  <= S_PROC_ADPAD0;
                    end if;
                end if;

            when S_PROC_ADPAD1 =>
                ca_set      <= '1';
                cb_set      <= '1';
                sel_M		<= "011";
                en_StateReg <= '1';
                count_next  <= count_r + 1;
                state_next  <= S_PROC_ADPAD0;

            when S_PROC_ADPAD0 =>
                sel_M		<= "100";
                en_StateReg <= '1';
                if (count_r <= PdivM-1) then
                    ca_set      <= '1';
                    cb_set      <= '1';
                elsif (count_r <= (PdivM_X2)-1) then
                    cb_set      <= '1';
                end if;
                if (count_r = (PdivM_X2)-1) then
                    count_next  <= (others => '0');
                    if (eoi_r = '1') then
                        state_next      <= S_PROC_DATPAD1;
                    else
                        state_next      <= S_PROC_PT;
                    end if;
                else
                    count_next  <= count_r + 1;
                end if;

            when S_PROC_PT =>
                if (bdo_ready = '1') and (bdi_valid = '1') then
                    bdo_valid   <= '1';
                    bdi_ready   <= '1';
                    ca_set      <= '1';
                    sel_M		<= "010";
                    sel_decrypt <= decrypt_r;
                    en_StateReg <= '1';
                    if (bdi_eot = '1') then
                        state_next  <= S_PROC_DATPAD1;
                        end_of_block <= '1';
                    end if;
                end if;

            when S_PROC_DATPAD1 =>
                ca_set      <= '1';
                sel_M		<= "011";
                en_StateReg <= '1';
                count_next  <= count_r + 1;
                state_next      <= S_PROC_DATPAD0;

            when S_PROC_DATPAD0 =>
                sel_M		<= "100";
                en_StateReg <= '1';
                if (count_r <= PdivM-1) then
                    ca_set      <= '1';
                elsif (count_r <= (PdivM_X2)-1) then
                    -- Noting
                end if;
                if (count_r = (PdivM_X2)-1) then
                    count_next  <= (others => '0');
                    state_next      <= S_FINAL;
                else
                    count_next  <= count_r + 1;
                end if;

            when S_FINAL =>
                ca_set      <= '1';
                cb_set      <= '1';
                is_final    <= '1';
                sel_M		<= "100";
                if (count_r < FINAL_COUNT) then
                    if (count_r >= (FINAL_COUNT- PdivM)) then
                        if (decrypt_r = '1') then
                            if (bdi_valid = '1') then
                                bdi_ready   <= '1';
                                en_StateReg <= '1';
                                count_next  <= count_r + 1;
                                if (tag_match = '0') then
                                    auth_fail_next  <= '1';
                                end if;
                                if (count_r = FINAL_COUNT-1) then
                                    if (msg_auth_ready = '1')then
                                        msg_auth_valid  <= '1';
                                        if (tag_match = '0') or (auth_fail_r = '1') then
                                            msg_auth   <= '0';
                                        else
                                            msg_auth   <= '1';
                                        end if;
                                        state_next  <= S_RESET;
                                    else
                                        state_next  <= S_WAIT_MSG_AUTH;
                                    end if;
                                end if;
                            end if;
                        else
                            if (bdo_ready = '1') then
                                bdo_valid   <= '1';
                                en_StateReg <= '1';
                                count_next  <= count_r + 1;
                                if (count_r = FINAL_COUNT-1) then
                                    end_of_block <= '1';
                                    state_next  <= S_RESET;
                                end if;
                            end if;
                        end if;
                    else
                        en_StateReg <= '1';
                        count_next  <= count_r + 1;
                    end if;
                end if;

            when S_WAIT_MSG_AUTH =>
                if (msg_auth_ready = '1')then
                    msg_auth_valid  <= '1';
                    if (auth_fail_r = '0') then
                        msg_auth   <= '1';
                    end if;
                    state_next  <= S_RESET;
                end if;

        end case;

    end process;

end behavioral;
